"""Edge port product block.

Edge port sets the boundary between Geant network and an external entity that could also be a different technological
domain still managed by GEANT. In other words, an Edge port determines where the network ends.
"""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle, strEnum

from gso.products.product_blocks.router import RouterBlock, RouterBlockInactive, RouterBlockProvisioning
from gso.utils.types.interfaces import LAGMemberList, PhysicalPortCapacity


class EncapsulationType(strEnum):
    """Types of encapsulation for edge ports."""

    DOT1Q = "dot1q"
    """Dot1Q supports multiple services for one customer or services for multiple customers."""
    QINQ = "qinq"
    """QinQ expands VLAN space by double-tagging frames."""
    NULL = "null"
    """Null supports a single service on the port."""


class EdgePortType(strEnum):
    """Types of edge ports."""

    CUSTOMER = "CUSTOMER"
    INFRASTRUCTURE = "INFRASTRUCTURE"
    PRIVATE = "PRIVATE"
    PUBLIC = "PUBLIC"
    RE_INTERCONNECT = "RE_INTERCONNECT"


class EdgePortAEMemberBlockInactive(
    ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="EdgePortAEMemberBlock"
):
    """An inactive Edge Port AE interface."""

    interface_name: str | None = None
    interface_description: str | None = None


class EdgePortAEMemberBlockProvisioning(EdgePortAEMemberBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A provisional Edge Port AE interface."""

    interface_name: str
    interface_description: str | None = None


class EdgePortAEMemberBlock(EdgePortAEMemberBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """An Edge Port AE interface."""

    interface_name: str
    interface_description: str | None = None


class EdgePortBlockInactive(
    ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="EdgePortBlock"
):
    """An edge port that's currently inactive. See ``EdgePortBlock``."""

    node: RouterBlockInactive | None = None
    edge_port_name: str | None = None
    edge_port_description: str | None = None
    enable_lacp: bool | None = None
    encapsulation: EncapsulationType = EncapsulationType.DOT1Q
    mac_address: str | None = None
    member_speed: PhysicalPortCapacity | None = None
    minimum_links: int | None = None
    edge_port_type: EdgePortType | None = None
    ignore_if_down: bool = False
    ga_id: str | None = None
    edge_port_ae_members: LAGMemberList[EdgePortAEMemberBlockInactive]


class EdgePortBlockProvisioning(EdgePortBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """An edge port that's being provisioned. See ``EdgePortBlock``."""

    node: RouterBlockProvisioning
    edge_port_name: str
    edge_port_description: str | None = None
    enable_lacp: bool
    encapsulation: EncapsulationType = EncapsulationType.DOT1Q
    mac_address: str | None = None
    member_speed: PhysicalPortCapacity
    minimum_links: int | None = None
    edge_port_type: EdgePortType
    ignore_if_down: bool = False
    ga_id: str | None = None
    edge_port_ae_members: LAGMemberList[EdgePortAEMemberBlockProvisioning]  # type: ignore[assignment]


class EdgePortBlock(EdgePortBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """An edge port that's currently deployed in the network.

    Attributes:
        node: The router that this Edge Port is connected to.
        edge_port_name: The name of the edge port, in our case, corresponds to the name of the LAG interface.
        edge_port_description: A description of the edge port.
        enable_lacp: Indicates whether LACP is enabled for this edge port.
        encapsulation: The type of encapsulation used on this edge port, by default DOT1Q.
        mac_address: The MAC address assigned to this edge port, if applicable.
        member_speed: The speed capacity of each member in the physical port.
        minimum_links: The minimum number of links required for this edge port.
        edge_port_type: The type of edge port (e.g., customer, private, public).
        ignore_if_down: If set to True, the edge port will be ignored if it is down.
        ga_id: The GEANT GA ID associated with this edge port, if any.
        edge_port_ae_members: A list of LAG members associated with this edge port.
    """

    node: RouterBlock
    edge_port_name: str
    edge_port_description: str | None = None
    enable_lacp: bool
    encapsulation: EncapsulationType = EncapsulationType.DOT1Q
    mac_address: str | None = None
    member_speed: PhysicalPortCapacity
    minimum_links: int | None = None
    edge_port_type: EdgePortType
    ignore_if_down: bool = False
    ga_id: str | None = None
    edge_port_ae_members: LAGMemberList[EdgePortAEMemberBlock]  # type: ignore[assignment]
