"""IP trunk product block that has all parameters of a subscription throughout its lifecycle."""

import ipaddress
from typing import Annotated

from annotated_types import Len
from orchestrator.domain.base import ProductBlockModel, T
from orchestrator.types import SubscriptionLifecycle, strEnum
from pydantic import AfterValidator
from pydantic_forms.validators import validate_unique_list
from typing_extensions import Doc

from gso.products.product_blocks.router import (
    RouterBlock,
    RouterBlockInactive,
    RouterBlockProvisioning,
)
from gso.utils.types.interfaces import LAGMemberList, PhysicalPortCapacity


class IptrunkType(strEnum):
    """Types of IP trunks. Can be dark fiber or a leased line."""

    DARK_FIBER = "Dark_fiber"
    LEASED = "Leased"


IptrunkSides = Annotated[
    list[T],
    AfterValidator(validate_unique_list),
    Len(min_length=2, max_length=2),
    Doc("A list of IP trunk interfaces that make up one side of a link."),
]


class IptrunkInterfaceBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="IptrunkInterfaceBlock",
):
    """An inactive IP trunk interface."""

    interface_name: str | None = None
    interface_description: str | None = None


class IptrunkInterfaceBlockProvisioning(IptrunkInterfaceBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """An IP trunk interface that is being provisioned."""

    interface_name: str
    interface_description: str | None = None


class IptrunkInterfaceBlock(IptrunkInterfaceBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """An active IP trunk interface."""

    interface_name: str
    interface_description: str | None = None


class IptrunkSideBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="IptrunkSideBlock",
):
    """An inactive IP trunk side."""

    iptrunk_side_node: RouterBlockInactive
    iptrunk_side_ae_iface: str | None = None
    ga_id: str | None = None
    iptrunk_side_ae_members: LAGMemberList[IptrunkInterfaceBlockInactive]


class IptrunkSideBlockProvisioning(IptrunkSideBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """An IP trunk side that is being provisioned."""

    iptrunk_side_node: RouterBlockProvisioning
    iptrunk_side_ae_iface: str
    ga_id: str | None = None
    iptrunk_side_ae_members: LAGMemberList[IptrunkInterfaceBlockProvisioning]  # type: ignore[assignment]


class IptrunkSideBlock(IptrunkSideBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """An active IP trunk side."""

    iptrunk_side_node: RouterBlock
    iptrunk_side_ae_iface: str
    ga_id: str | None = None
    iptrunk_side_ae_members: LAGMemberList[IptrunkInterfaceBlock]  # type: ignore[assignment]


class IptrunkBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="IptrunkBlock",
):
    """A trunk that's currently inactive, see `IptrunkBlock`."""

    gs_id: str | None = None
    iptrunk_description: str | None = None
    iptrunk_type: IptrunkType | None = None
    iptrunk_speed: PhysicalPortCapacity | None = None
    iptrunk_minimum_links: int | None = None
    iptrunk_isis_metric: int | None = None
    iptrunk_ipv4_network: ipaddress.IPv4Network | None = None
    iptrunk_ipv6_network: ipaddress.IPv6Network | None = None
    iptrunk_sides: IptrunkSides[IptrunkSideBlockInactive]


class IptrunkBlockProvisioning(IptrunkBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A trunk that's currently being provisioned, see `IptrunkBlock`."""

    gs_id: str | None = None
    iptrunk_description: str | None = None
    iptrunk_type: IptrunkType | None = None
    iptrunk_speed: PhysicalPortCapacity | None = None
    iptrunk_minimum_links: int | None = None
    iptrunk_isis_metric: int | None = None
    iptrunk_ipv4_network: ipaddress.IPv4Network | None = None
    iptrunk_ipv6_network: ipaddress.IPv6Network | None = None
    iptrunk_sides: IptrunkSides[IptrunkSideBlockProvisioning]  # type: ignore[assignment]


class IptrunkBlock(IptrunkBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A trunk that's currently deployed in the network.

    Attributes:
        gs_id: GÉANT service ID associated with this trunk.
        iptrunk_description: A human-readable description of this trunk.
        iptrunk_type: The type of trunk, can be either dark fibre or leased capacity.
        iptrunk_speed: The speed of the trunk, measured per interface associated with it.
        iptrunk_minimum_links: The minimum amount of links the trunk should consist of.
        iptrunk_isis_metric: The ISIS metric of this link.
        iptrunk_ipv4_network: The IPv4 network used for this trunk.
        iptrunk_ipv6_network: The IPv6 network used for this trunk.
        iptrunk_sides: The two sides that the trunk is connected to.
    """

    gs_id: str | None = None
    iptrunk_description: str | None = None
    iptrunk_type: IptrunkType
    iptrunk_speed: PhysicalPortCapacity
    iptrunk_minimum_links: int
    iptrunk_isis_metric: int
    iptrunk_ipv4_network: ipaddress.IPv4Network
    iptrunk_ipv6_network: ipaddress.IPv6Network
    iptrunk_sides: IptrunkSides[IptrunkSideBlock]  # type: ignore[assignment]
