"""Layer 2 Circuit product block."""

from collections.abc import Sequence
from typing import Annotated, TypeVar

from annotated_types import Len
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from pydantic import AfterValidator
from pydantic_forms.types import strEnum
from pydantic_forms.validators import validate_unique_list
from typing_extensions import Doc

from gso.products.product_blocks.service_binding_port import (
    ServiceBindingPort,
    ServiceBindingPortInactive,
    ServiceBindingPortProvisioning,
)
from gso.utils.types.interfaces import BandwidthString
from gso.utils.types.virtual_identifiers import VC_ID, VLAN_ID


class Layer2CircuitSideBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="Layer2CircuitSideBlock",
):
    """One inactive side of a Layer 2 Circuit."""

    sbp: ServiceBindingPortInactive


class Layer2CircuitSideBlockProvisioning(
    Layer2CircuitSideBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    """One provisioning side of a Layer 2 Circuit."""

    sbp: ServiceBindingPortProvisioning


class Layer2CircuitSideBlock(Layer2CircuitSideBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """One side of a Layer 2 Circuit."""

    sbp: ServiceBindingPort


Layer2CircuitSideBlockType = TypeVar(
    "Layer2CircuitSideBlockType",
    "Layer2CircuitSideBlockInactive",
    "Layer2CircuitSideBlockProvisioning",
    "Layer2CircuitSideBlock",
)

Layer2CircuitSides = Annotated[
    Sequence[Layer2CircuitSideBlockType],
    AfterValidator(validate_unique_list),
    Len(min_length=2, max_length=2),
    Doc("A list of two Layer 2 Circuit sides."),
]


class Layer2CircuitType(strEnum):
    """The two types of Layer 2 Circuit."""

    TAGGED = "TAGGED"
    UNTAGGED = "UNTAGGED"


class Layer2CircuitBlockInactive(
    ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="Layer2CircuitBlock"
):
    """An inactive Layer 2 Circuit, see ``Layer2CircuitBlock``."""

    layer_2_circuit_sides: Layer2CircuitSides[Layer2CircuitSideBlockInactive]
    virtual_circuit_id: VC_ID | None = None
    layer_2_circuit_type: Layer2CircuitType | None = None
    vlan_range_lower_bound: VLAN_ID | None = None
    vlan_range_upper_bound: VLAN_ID | None = None
    policer_enabled: bool | None = None
    policer_burst_rate: BandwidthString | None = None
    bandwidth: BandwidthString | None = None


class Layer2CircuitBlockProvisioning(Layer2CircuitBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A provisioning Layer 2 Circuit, see ``Layer2CircuitBlock``."""

    layer_2_circuit_sides: Layer2CircuitSides[Layer2CircuitSideBlockProvisioning]
    virtual_circuit_id: VC_ID
    layer_2_circuit_type: Layer2CircuitType
    vlan_range_lower_bound: VLAN_ID | None
    vlan_range_upper_bound: VLAN_ID | None
    policer_enabled: bool
    policer_burst_rate: BandwidthString | None
    bandwidth: BandwidthString | None


class Layer2CircuitBlock(Layer2CircuitBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """An active Layer 2 Circuit.

    Attributes:
        layer_2_circuit_sides: The two sides that the Layer 2 Circuit is connected to.
        virtual_circuit_id: Virtual Circuit ID of this Layer 2 Circuit.
        layer_2_circuit_type: The type of circuit, can be tagged or untagged.
        vlan_range_lower_bound: If tagged, the lower and upper bounds will set the VLAN range.
        vlan_range_upper_bound: Lower and Upper bounds are including.
        policer_enabled: Whether this Layer 2 Circuit is policed.
        policer_burst_rate: If policed, the burst rate of the policer.
        bandwidth: If policed, the bandwidth of the policer is stored.
    """

    layer_2_circuit_sides: Layer2CircuitSides[Layer2CircuitSideBlock]
    virtual_circuit_id: VC_ID
    layer_2_circuit_type: Layer2CircuitType
    vlan_range_lower_bound: VLAN_ID | None
    vlan_range_upper_bound: VLAN_ID | None
    policer_enabled: bool
    policer_burst_rate: BandwidthString | None
    bandwidth: BandwidthString | None
