"""Product block for office router products."""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle

from gso.products.product_blocks.site import (
    SiteBlock,
    SiteBlockInactive,
    SiteBlockProvisioning,
)
from gso.utils.shared_enums import Vendor
from gso.utils.types.ip_address import IPv4AddressType, IPv6AddressType, PortNumber


class OfficeRouterBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="OfficeRouterBlock",
):
    """An office router that's being currently inactive. See `OfficeRouterBlock`."""

    office_router_fqdn: str | None = None
    office_router_ts_port: PortNumber | None = None
    office_router_lo_ipv4_address: IPv4AddressType | None = None
    office_router_lo_ipv6_address: IPv6AddressType | None = None
    office_router_site: SiteBlockInactive | None
    vendor: Vendor | None = Vendor.JUNIPER


class OfficeRouterBlockProvisioning(OfficeRouterBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """An office router that's being provisioned. See `RouterBlock`."""

    office_router_fqdn: str | None
    office_router_ts_port: PortNumber | None
    office_router_lo_ipv4_address: IPv4AddressType | None
    office_router_lo_ipv6_address: IPv6AddressType | None
    office_router_site: SiteBlockProvisioning | None
    vendor: Vendor


class OfficeRouterBlock(OfficeRouterBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """An office router that's currently deployed in the network.

    Attributes:
        office_router_fqdn: Office router FQDN.
        office_router_ts_port: The port of the terminal server that this office router is connected to. Used to offer
            out of band access.
        office_router_lo_ipv4_address: The IPv4 loopback address of the office router.
        office_router_lo_ipv6_address: The IPv6 loopback address of the office router.
        office_router_site: The `Site` that this office router resides in. Both physically and computationally.
        vendor: The vendor of an office router. Defaults to Juniper.
    """

    office_router_fqdn: str
    office_router_ts_port: PortNumber
    office_router_lo_ipv4_address: IPv4AddressType
    office_router_lo_ipv6_address: IPv6AddressType
    office_router_site: SiteBlock
    vendor: Vendor
