"""Product block for `Router` products."""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle, strEnum

from gso.products.product_blocks.site import (
    SiteBlock,
    SiteBlockInactive,
    SiteBlockProvisioning,
)
from gso.utils.shared_enums import Vendor
from gso.utils.types.ip_address import IPv4AddressType, IPv6AddressType, PortNumber


class RouterRole(strEnum):
    """Enumerator for the different types of routers."""

    P = "p"
    PE = "pe"
    AMT = "amt"


class RouterBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="RouterBlock",
):
    """A router that's being currently inactive. See `RouterBlock`."""

    router_fqdn: str | None = None
    router_ts_port: PortNumber | None = None
    router_access_via_ts: bool | None = None
    router_lo_ipv4_address: IPv4AddressType | None = None
    router_lo_ipv6_address: IPv6AddressType | None = None
    router_lo_iso_address: str | None = None
    router_role: RouterRole | None = None
    router_site: SiteBlockInactive | None
    vendor: Vendor | None = None


class RouterBlockProvisioning(RouterBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A router that's being provisioned. See `RouterBlock`."""

    router_fqdn: str
    router_ts_port: PortNumber
    router_access_via_ts: bool
    router_lo_ipv4_address: IPv4AddressType
    router_lo_ipv6_address: IPv6AddressType
    router_lo_iso_address: str
    router_role: RouterRole
    router_site: SiteBlockProvisioning
    vendor: Vendor


class RouterBlock(RouterBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A router that's currently deployed in the network.

    Attributes:
        router_fqdn: FQDN of a router.
        router_ts_port: The port of the terminal server that this router is connected to. Used to offer out of band
            access.
        router_access_via_ts: Whether this router should be accessed through the terminal server, or through its
            loopback address.
        router_lo_ipv4_address: The IPv4 loopback address of the router.
        router_lo_ipv6_address: The IPv6 loopback address of the router.
        router_lo_iso_address: The ISO NET of the router, used for ISIS support.
        router_role: The role of the router, which can be any of the values defined in `RouterRole`.
        router_site: The site that this router resides in. Both physically and computationally.
        vendor: The vendor of a router.
    """

    router_fqdn: str
    router_ts_port: PortNumber
    router_access_via_ts: bool
    router_lo_ipv4_address: IPv4AddressType
    router_lo_ipv6_address: IPv6AddressType
    router_lo_iso_address: str
    router_role: RouterRole
    router_site: SiteBlock
    vendor: Vendor
