"""Service Binding Port.

A service binding port is used to logically attach an edge port to a customer service using a VLAN.
"""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from pydantic import Field

from gso.products.product_blocks.bgp_session import (
    BGPSession,
    BGPSessionInactive,
    BGPSessionProvisioning,
)
from gso.products.product_blocks.edge_port import EdgePortBlock, EdgePortBlockInactive, EdgePortBlockProvisioning
from gso.utils.shared_enums import SBPType
from gso.utils.types.ip_address import IPv4AddressType, IPv4Netmask, IPv6AddressType, IPv6Netmask
from gso.utils.types.virtual_identifiers import VLAN_ID


class BFDSettingsInactive(
    ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="BFDSettings"
):
    """Settings for BFD, see `BFDSettings`."""

    bfd_enabled: bool = False
    bfd_interval_rx: int | None = None
    bfd_interval_tx: int | None = None
    bfd_multiplier: int | None = None


class BFDSettingsProvisioning(BFDSettingsInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """Settings for BFD, see `BFDSettings`."""

    bfd_enabled: bool
    bfd_interval_rx: int | None = None
    bfd_interval_tx: int | None = None
    bfd_multiplier: int | None


class BFDSettings(BFDSettingsProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A set of settings for BFD.

    Attributes:
        bfd_enabled: Whether BFD is enabled.
        bfd_interval_rx: The BFD interval RX, if enabled.
        bfd_interval_tx: The BFD interval TX, if enabled.
        bfd_multiplier: The BFD multiplier, if enabled.
    """

    bfd_enabled: bool
    bfd_interval_rx: int | None
    bfd_interval_tx: int | None
    bfd_multiplier: int | None


class ServiceBindingPortInactive(
    ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="ServiceBindingPort"
):
    """A Service Binding Port that's currently inactive. See `ServiceBindingPort`."""

    is_tagged: bool | None = None
    vlan_id: VLAN_ID | None = None
    sbp_type: SBPType | None = None
    ipv4_address: IPv4AddressType | None = None
    ipv4_mask: IPv4Netmask | None = None
    ipv6_address: IPv6AddressType | None = None
    ipv6_mask: IPv6Netmask | None = None
    custom_firewall_filters: bool | None = None
    gs_id: str | None = None
    bgp_session_list: list[BGPSessionInactive] = Field(default_factory=list)
    edge_port: EdgePortBlockInactive | None = None
    v4_bfd_settings: BFDSettingsInactive
    v6_bfd_settings: BFDSettingsInactive


class ServiceBindingPortProvisioning(ServiceBindingPortInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A Service Binding Port that's currently being provisioned. See `ServiceBindingPort`."""

    is_tagged: bool
    vlan_id: VLAN_ID | None = None
    sbp_type: SBPType
    ipv4_address: IPv4AddressType | None = None
    ipv4_mask: IPv4Netmask | None = None
    ipv6_address: IPv6AddressType | None = None
    ipv6_mask: IPv6Netmask | None = None
    custom_firewall_filters: bool
    gs_id: str
    bgp_session_list: list[BGPSessionProvisioning]  # type: ignore[assignment]
    edge_port: EdgePortBlockProvisioning
    v4_bfd_settings: BFDSettingsProvisioning
    v6_bfd_settings: BFDSettingsProvisioning


class ServiceBindingPort(ServiceBindingPortProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A service binding port that is actively used in the network.

    Attributes:
        is_tagged: Whether this VLAN is tagged or not.
        vlan_id: The VLAN ID.
        sbp_type: Is this service binding port layer 2 or 3?
        ipv4_address: If layer 3, IPv4 resources.
        ipv4_mask: IPv4 subnet mask.
        ipv6_address: If layer 3, IPv6 resources.
        ipv6_mask: IPv6 subnet mask.
        custom_firewall_filters: Any custom firewall filters that the partner may require.
        gs_id: The GÉANT service ID of this binding port.
        bgp_session_list: The BGP sessions associated with this service binding port.
        edge_port: The Edge Port on which this SBP resides.
        v4_bfd_settings: BFD settings for IPv4.
        v6_bfd_settings: BFD settings for IPv6.
    """

    is_tagged: bool
    vlan_id: VLAN_ID | None = None
    sbp_type: SBPType
    ipv4_address: IPv4AddressType | None = None
    ipv4_mask: IPv4Netmask | None = None
    ipv6_address: IPv6AddressType | None = None
    ipv6_mask: IPv6Netmask | None = None
    custom_firewall_filters: bool
    gs_id: str
    bgp_session_list: list[BGPSession]  # type: ignore[assignment]
    edge_port: EdgePortBlock
    v4_bfd_settings: BFDSettings
    v6_bfd_settings: BFDSettings
