"""The product block that describes a site subscription."""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle, strEnum

from gso.utils.types.coordinates import LatitudeCoordinate, LongitudeCoordinate
from gso.utils.types.ip_address import IPAddress
from gso.utils.types.site_name import SiteName


class SiteTier(strEnum):
    """The tier of a site, ranging from 1 to 4.

    A lower value corresponds to a larger amount of installed equipment, and a higher cardinality of the connectivity to
    and from a site.
    """

    TIER1 = 1
    TIER2 = 2
    TIER3 = 3
    TIER4 = 4


class SiteBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="SiteBlock",
):
    """A site that's currently inactive, see `SiteBlock`."""

    site_name: SiteName | None = None
    site_city: str | None = None
    site_country: str | None = None
    site_country_code: str | None = None
    site_latitude: LatitudeCoordinate | None = None
    site_longitude: LongitudeCoordinate | None = None
    site_internal_id: int | None = None
    site_bgp_community_id: int | None = None
    site_tier: SiteTier | None = None
    site_ts_address: IPAddress | None = None
    site_contains_optical_equipment: bool | None = None


class SiteBlockProvisioning(SiteBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A site that's currently being provisioned, see `SiteBlock`."""

    site_name: SiteName
    site_city: str
    site_country: str
    site_country_code: str
    site_latitude: LatitudeCoordinate
    site_longitude: LongitudeCoordinate
    site_internal_id: int
    site_bgp_community_id: int
    site_tier: SiteTier
    site_ts_address: IPAddress
    site_contains_optical_equipment: bool


class SiteBlock(SiteBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A site that's currently available for routers and services to be hosted at.

    Attributes:
        site_name: The name of the site, that will dictate part of the FQDN of routers that are hosted at this site. For
            example: `router.X.Y.geant.net`, where X denotes the name of the site.
        site_city: The city at which the site is located.
        site_country: The country in which the site is located.
        site_country_code: The code of the corresponding country. This is also used for the FQDN, following the example
            given for the site name, the country code would end up in the Y position.
        site_latitude: The latitude of the site, used for SNMP purposes.
        site_longitude: Similar to the latitude, the longitude of a site.
        site_internal_id: The internal ID used within GÉANT to denote a site.
        site_bgp_community_id: The BGP community ID of a site, used to advertise routes learned at this site.
        site_tier: The tier of a site, as described in `SiteTier`.
        site_ts_address: The address of the terminal server that this router is connected to. The terminal server
            provides out of band access. This is required in case a link goes down, or when a router is initially added
            to the network, and it does not have any IP trunks connected to it.
        site_contains_optical_equipment: Whether this site contains optical equipment, which dictates the need for a DCN
            management VLAN.
    """

    site_name: SiteName
    site_city: str
    site_country: str
    site_country_code: str
    site_latitude: LatitudeCoordinate
    site_longitude: LongitudeCoordinate
    site_internal_id: int
    site_bgp_community_id: int
    site_tier: SiteTier
    site_ts_address: IPAddress
    site_contains_optical_equipment: bool
