"""Product block for Super PoP Switch products."""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle

from gso.products.product_blocks.site import (
    SiteBlock,
    SiteBlockInactive,
    SiteBlockProvisioning,
)
from gso.utils.shared_enums import Vendor
from gso.utils.types.ip_address import IPv4AddressType, PortNumber


class SuperPopSwitchBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="SuperPopSwitchBlock",
):
    """A Super PoP switch that's being currently inactive. See `SuperPoPSwitchBlock`."""

    super_pop_switch_fqdn: str | None = None
    super_pop_switch_ts_port: PortNumber | None = None
    super_pop_switch_mgmt_ipv4_address: IPv4AddressType | None = None
    super_pop_switch_site: SiteBlockInactive | None
    vendor: Vendor | None = Vendor.JUNIPER


class SuperPopSwitchBlockProvisioning(SuperPopSwitchBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A Super PoP switch that's being provisioned. See `SuperPoPSwitchBlock`."""

    super_pop_switch_fqdn: str | None
    super_pop_switch_ts_port: PortNumber | None
    super_pop_switch_mgmt_ipv4_address: IPv4AddressType | None
    super_pop_switch_site: SiteBlockProvisioning | None
    vendor: Vendor | None


class SuperPopSwitchBlock(SuperPopSwitchBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A Super PoP switch that's currently deployed in the network.

    Attributes:
        super_pop_switch_fqdn: Super PoP switch FQDN.
        super_pop_switch_ts_port: The port of the terminal server that this Super PoP switch is connected to. Used to
            offer out of band access.
        super_pop_switch_mgmt_ipv4_address: The IPv4 management address of the Super PoP switch.
        super_pop_switch_site: The `Site` that this Super PoP switch resides in. Both physically and computationally.
        vendor: The vendor of a Super PoP switch. Defaults to Juniper.
    """

    super_pop_switch_fqdn: str
    super_pop_switch_ts_port: PortNumber
    super_pop_switch_mgmt_ipv4_address: IPv4AddressType
    super_pop_switch_site: SiteBlock
    vendor: Vendor
