"""A task for checking site connectivity.

When a new router is planned to be deployed, it is good practice to verify that the OOB connectivity is functioning
correctly. This task takes a site and a port number as input, and checks whether there is reachability across the OOB
access.
"""

from orchestrator import workflow
from orchestrator.forms import SubmitFormPage
from orchestrator.targets import Target
from orchestrator.workflow import StepList, begin, done, step
from pydantic import ConfigDict
from pydantic_forms.types import FormGenerator, UUIDstr

from gso.products.product_types.site import Site
from gso.services.lso_client import LSOState, lso_interaction
from gso.utils.helpers import active_site_selector
from gso.utils.types.ip_address import PortNumber


def _initial_input_form_generator() -> FormGenerator:
    class CheckSiteConnectivityForm(SubmitFormPage):
        model_config = ConfigDict(title="Verify NETCONF connectivity to a Site")

        site: active_site_selector()  # type: ignore[valid-type]
        port: PortNumber

    user_input = yield CheckSiteConnectivityForm
    return user_input.model_dump()


@step("Check NETCONF connectivity")
def check_netconf_connectivity(site: UUIDstr, port: PortNumber) -> LSOState:
    """Run an Ansible playbook that validates NETCONF connectivity to a Site."""
    site_subscription = Site.from_subscription(site).site
    return {
        "playbook_name": "gap_ansible/playbooks/check_netconf_connectivity.yaml",
        "extra_vars": {"port_number": port},
        "inventory": {"all": {"hosts": {site_subscription.site_ts_address: None}}},
    }


@workflow("Check Site Connectivity", _initial_input_form_generator, Target.SYSTEM)
def task_check_site_connectivity() -> StepList:
    """Check successful NETCONF connectivity of a Site."""
    return begin >> lso_interaction(check_netconf_connectivity) >> done
