from unittest.mock import patch

import pytest
from pydantic_forms.exceptions import FormValidationError

from gso.products.product_blocks.iptrunk import IptrunkType
from gso.products.product_types.iptrunk import Iptrunk
from gso.utils.shared_enums import Vendor
from gso.utils.types.interfaces import LAGMemberList, PhysicalPortCapacity
from test.conftest import UseJuniperSide
from test.workflows import (
    assert_complete,
    assert_lso_interaction_success,
    extract_state,
    run_workflow,
)
from test.workflows.iptrunk.test_create_iptrunk import MockedNetboxClient


@pytest.fixture()
def input_form_iptrunk_data(
    request,
    faker,
    iptrunk_subscription_factory,
    router_subscription_factory,
    iptrunk_side_subscription_factory,
):
    use_juniper = getattr(request, "param", UseJuniperSide.NONE)
    if use_juniper == UseJuniperSide.SIDE_A:
        side_node = router_subscription_factory(vendor=Vendor.JUNIPER)
        side_a_node = iptrunk_side_subscription_factory(iptrunk_side_node=side_node)
        side_b_node = iptrunk_side_subscription_factory()
        new_side_a_ae_members = faker.link_members_juniper()[0:2]
        new_side_b_ae_members = faker.link_members_nokia()[0:2]
    elif use_juniper == UseJuniperSide.SIDE_B:
        side_node = router_subscription_factory(vendor=Vendor.JUNIPER)
        side_a_node = iptrunk_side_subscription_factory()
        side_b_node = iptrunk_side_subscription_factory(iptrunk_side_node=side_node)
        new_side_a_ae_members = faker.link_members_nokia()[0:2]
        new_side_b_ae_members = faker.link_members_juniper()[0:2]
    elif use_juniper == UseJuniperSide.SIDE_BOTH:
        side_node_1 = router_subscription_factory(vendor=Vendor.JUNIPER)
        side_node_2 = router_subscription_factory(vendor=Vendor.JUNIPER)
        side_a_node = iptrunk_side_subscription_factory(iptrunk_side_node=side_node_1)
        side_b_node = iptrunk_side_subscription_factory(iptrunk_side_node=side_node_2)
        new_side_a_ae_members = faker.link_members_juniper()[0:2]
        new_side_b_ae_members = faker.link_members_juniper()[0:2]
    else:
        side_a_node = iptrunk_side_subscription_factory()
        side_b_node = iptrunk_side_subscription_factory()
        new_side_a_ae_members = faker.link_members_nokia()[0:2]
        new_side_b_ae_members = faker.link_members_nokia()[0:2]

    product_id = iptrunk_subscription_factory(iptrunk_sides=[side_a_node, side_b_node])

    new_sid = faker.gs_id()
    new_description = faker.sentence()
    new_type = IptrunkType.LEASED
    new_speed = PhysicalPortCapacity.FOUR_HUNDRED_GIGABIT_PER_SECOND
    new_link_count = 2

    new_side_a_gid = faker.ga_id()

    new_side_b_gid = faker.ga_id()

    return [
        {"subscription_id": product_id},
        {
            "tt_number": faker.tt_number(),
            "gs_id": new_sid,
            "iptrunk_description": new_description,
            "iptrunk_type": new_type,
            "iptrunk_speed": new_speed,
            "iptrunk_number_of_members": new_link_count,
        },
        {},
        {
            "side_a_ga_id": new_side_a_gid,
            "side_a_ae_members": new_side_a_ae_members,
        },
        {
            "side_b_ga_id": new_side_b_gid,
            "side_b_ae_members": new_side_b_ae_members,
        },
    ]


@pytest.mark.parametrize(
    "input_form_iptrunk_data",
    [UseJuniperSide.NONE, UseJuniperSide.SIDE_A, UseJuniperSide.SIDE_B, UseJuniperSide.SIDE_BOTH],
    indirect=True,
)
@pytest.mark.workflow()
@patch("gso.services.lso_client._send_request")
@patch("gso.services.netbox_client.NetboxClient.get_available_interfaces")
@patch("gso.services.netbox_client.NetboxClient.attach_interface_to_lag")
@patch("gso.services.netbox_client.NetboxClient.reserve_interface")
@patch("gso.services.netbox_client.NetboxClient.allocate_interface")
@patch("gso.services.netbox_client.NetboxClient.free_interface")
@patch("gso.services.netbox_client.NetboxClient.detach_interfaces_from_lag")
def test_iptrunk_modify_trunk_interface_success(
    mocked_detach_interfaces_from_lag,
    mocked_free_interface,
    mocked_allocate_interface,
    mocked_reserve_interface,
    mocked_attach_interface_to_lag,
    mocked_get_available_interfaces,
    mock_provision_ip_trunk,
    input_form_iptrunk_data,
    faker,
    data_config_filename,
):
    #  Set up mock return values
    mocked_netbox = MockedNetboxClient()
    mocked_get_available_interfaces.return_value = mocked_netbox.get_available_interfaces()
    mocked_attach_interface_to_lag.return_value = mocked_netbox.attach_interface_to_lag()
    mocked_reserve_interface.return_value = mocked_netbox.reserve_interface()
    mocked_allocate_interface.return_value = mocked_netbox.allocate_interface()
    mocked_free_interface.return_value = mocked_netbox.free_interface()
    mocked_detach_interfaces_from_lag.return_value = mocked_netbox.detach_interfaces_from_lag()

    #  Run workflow
    result, process_stat, step_log = run_workflow("modify_trunk_interface", input_form_iptrunk_data)
    state = extract_state(result)
    lso_interaction_count = 10 if state["capacity_has_changed"] else 2

    for _ in range(lso_interaction_count):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_complete(result)

    state = extract_state(result)
    subscription_id = state["subscription_id"]
    subscription = Iptrunk.from_subscription(subscription_id)

    assert subscription.status == "active"
    assert mock_provision_ip_trunk.call_count == lso_interaction_count
    # Assert all Netbox calls have been made
    new_sid = input_form_iptrunk_data[1]["gs_id"]
    new_side_a_gid = input_form_iptrunk_data[3]["side_a_ga_id"]
    new_side_a_ae_members = input_form_iptrunk_data[3]["side_a_ae_members"]
    new_side_b_gid = input_form_iptrunk_data[4]["side_b_ga_id"]
    new_side_b_ae_members = input_form_iptrunk_data[4]["side_b_ae_members"]

    # Only Nokia interfaces are checked
    vendor_side_a = subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.vendor
    vendor_side_b = subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.vendor
    num_ifaces = (len(new_side_a_ae_members) if vendor_side_a == Vendor.NOKIA else 0) + (
        len(new_side_b_ae_members) if vendor_side_b == Vendor.NOKIA else 0
    )

    # Define free interfaces for only nokia sides
    num_free_ifaces = 2 * (vendor_side_a == Vendor.NOKIA) + 2 * (vendor_side_b == Vendor.NOKIA)

    # LAG interface for nokia sides
    num_lag_ifaces = int(vendor_side_a == Vendor.NOKIA) + int(vendor_side_b == Vendor.NOKIA)

    assert mocked_reserve_interface.call_count == num_ifaces  # Only nokia interfaces per side num is randomly generated
    assert mocked_attach_interface_to_lag.call_count == num_ifaces
    assert mocked_free_interface.call_count == num_free_ifaces  # 2 interfaces per nokia side(The old ones)
    assert mocked_detach_interfaces_from_lag.call_count == num_lag_ifaces  # 1 time per nokia side

    # Assert all subscription properties have been updated correctly
    side_names = sorted([
        subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_site.site_name,
        subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_site.site_name,
    ])
    assert subscription.description == f"IP trunk {side_names[0]} {side_names[1]}, {new_sid}"
    assert subscription.iptrunk.gs_id == input_form_iptrunk_data[1]["gs_id"]
    assert subscription.iptrunk.iptrunk_description == input_form_iptrunk_data[1]["iptrunk_description"]
    assert subscription.iptrunk.iptrunk_type == input_form_iptrunk_data[1]["iptrunk_type"]
    assert subscription.iptrunk.iptrunk_speed == input_form_iptrunk_data[1]["iptrunk_speed"]
    assert subscription.iptrunk.iptrunk_minimum_links == input_form_iptrunk_data[1]["iptrunk_number_of_members"] - 1
    assert subscription.iptrunk.iptrunk_sides[0].ga_id == new_side_a_gid

    def _find_interface_by_name(interfaces: LAGMemberList, name: str):
        for interface in interfaces:
            if interface.interface_name == name:
                return interface
        msg = f"Interface {name} not found!"
        raise IndexError(msg)

    for member in subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members:
        assert (
            member.interface_description
            == _find_interface_by_name(new_side_a_ae_members, member.interface_name).interface_description
        )

    assert subscription.iptrunk.iptrunk_sides[1].ga_id == new_side_b_gid

    for member in subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members:
        assert (
            member.interface_description
            == _find_interface_by_name(new_side_b_ae_members, member.interface_name).interface_description
        )


@pytest.mark.workflow()
@pytest.mark.parametrize("invalid_ga_id", ["GA-11111", "GS-1234", "GS_12345", "GS-100000"])
def test_modify_iptrunk_interface_with_invalid_ga_id(
    input_form_iptrunk_data, faker, invalid_ga_id, iptrunk_side_subscription_factory, iptrunk_subscription_factory
):
    input_data = input_form_iptrunk_data
    input_data[3]["side_a_ga_id"] = invalid_ga_id
    iptrunk_subscription_factory(
        iptrunk_sides=[iptrunk_side_subscription_factory(ga_id="GA-11111"), iptrunk_side_subscription_factory()]
    )

    #  Run workflow
    with pytest.raises(FormValidationError):
        run_workflow("modify_edge_port", input_data)
