from unittest.mock import patch

import pytest
from orchestrator.types import SubscriptionLifecycle

from gso.products import ProductName
from gso.products.product_types.lan_switch_interconnect import LanSwitchInterconnect
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.types.virtual_identifiers import DEFAULT_DCN_MANAGEMENT_VLAN_ID
from test.services.conftest import MockedNetboxClient
from test.workflows import assert_complete, extract_state, run_workflow


@pytest.fixture()
def _netbox_client_mock():
    # Mock NetboxClient methods
    with (
        patch("gso.services.netbox_client.NetboxClient.get_device_by_name") as mock_get_device_by_name,
        patch("gso.services.netbox_client.NetboxClient.get_available_interfaces") as mock_get_available_interfaces,
        patch("gso.services.netbox_client.NetboxClient.get_available_lags_in_range") as mock_available_lags_in_range,
    ):
        mock_get_device_by_name.return_value = MockedNetboxClient().get_device_by_name()
        mock_get_available_interfaces.return_value = MockedNetboxClient().get_available_interfaces()
        mock_available_lags_in_range.return_value = MockedNetboxClient().get_available_lags_in_range()

        yield


@pytest.fixture()
def input_form_data(faker, router_subscription_factory, switch_subscription_factory, site_subscription_factory):
    def _input_form_data(site_contains_optical_equipment):
        return [
            {
                "product": get_product_id_by_name(ProductName.LAN_SWITCH_INTERCONNECT),
            },
            {
                "tt_number": faker.tt_number(),
                "router_side": router_subscription_factory(
                    router_site=site_subscription_factory(
                        site_contains_optical_equipment=site_contains_optical_equipment
                    )
                ),
                "switch_side": switch_subscription_factory(),
                "description": faker.sentence(),
                "minimum_link_count": 2,
            },
            {
                "router_side_iface": "lag-1",
                "router_side_ae_members": faker.link_members_nokia()[:2],
            },
            {
                "switch_side_iface": faker.network_interface(),
                "switch_side_ae_members": faker.link_members_juniper()[:2],
            },
            {},
        ]

    return _input_form_data


@pytest.mark.workflow()
@pytest.mark.parametrize("site_contains_optical_equipment", [True, False])
@patch("gso.workflows.lan_switch_interconnect.create_lan_switch_interconnect.create_v6_network_by_ip")
@patch("gso.workflows.lan_switch_interconnect.create_lan_switch_interconnect.create_v4_network_by_ip")
@patch("gso.workflows.lan_switch_interconnect.create_lan_switch_interconnect.create_host_by_ip")
def test_create_lan_switch_interconnect_success(
    mock_create_host,
    mock_create_v4_network,
    mock_create_v6_network,
    site_contains_optical_equipment,
    input_form_data,
    _netbox_client_mock,  # noqa: PT019
):
    result, _, _ = run_workflow("create_lan_switch_interconnect", input_form_data(site_contains_optical_equipment))

    assert_complete(result)
    state = extract_state(result)
    subscription_id = state["subscription_id"]
    subscription = LanSwitchInterconnect.from_subscription(subscription_id)
    assert subscription.status == SubscriptionLifecycle.ACTIVE
    assert mock_create_v4_network.call_count == 1
    assert mock_create_v6_network.call_count == 1
    assert mock_create_host.call_count == 4
    assert subscription.lan_switch_interconnect.dcn_management_vlan_id == (
        DEFAULT_DCN_MANAGEMENT_VLAN_ID if site_contains_optical_equipment else None
    )
